#ifndef STACKLESS_IMPL_H
#define STACKLESS_IMPL_H

#include "Python.h"

#ifdef STACKLESS

#ifdef __cplusplus
extern "C" {
#endif

/* define this one to get extensive debugging */
#undef DEEP_DEBUG

/* support exports under Windows */
#ifdef USE_DL_IMPORT
#undef DL_IMPORT
#undef DL_EXPORT
#define DL_IMPORT(RTYPE) __declspec(dllimport) RTYPE
#define DL_EXPORT(RTYPE) __declspec(dllexport) RTYPE
#endif

/* keep this definition and the above crap until we stop supporting 2.2 */
#ifndef PyAPI_FUNC
#define PyAPI_FUNC(thing) extern DL_IMPORT(thing)
#endif


#include "structmember.h"
#include "compile.h"
#include "frameobject.h"

#include "flextype.h"
#include "stackless_frame.h"
#include "stackless_structs.h"
#include "nonrecursive.h"

#undef STACKLESS_SPY
/* 
 * if a platform wants to support self-inspection via _peek,
 * it must provide a function or macro CANNOT_READ_MEM(adr, len)
 * which allows to spy at memory without causing exceptions.
 * This would usually be done in place with the assembly macros.
 */

/* extract the lower bit of Py_TPFLAGS_HAVE_STACKLESS_EXTENSION */

#define Py_TPFLAGS_HAVE_NR_THINGS \
	(Py_TPFLAGS_HAVE_STACKLESS_EXTENSION & (Py_TPFLAGS_HAVE_STACKLESS_EXTENSION >> 1))

/********************************************************************
 *
 * This section defines/references stuff from stacklesseval.c
 *
 ********************************************************************/

/*** access to system-wide globals from stacklesseval.c ***/

PyAPI_DATA(int) slp_enable_softswitch;
PyAPI_DATA(int) slp_enable_autoschedule;

PyAPI_FUNC(int) slp_transfer(PyCStackObject **cstprev, PyCStackObject *cst, PyTaskletObject *prev);

PyAPI_FUNC(void) _PyStackless_InitTypes(void);
PyAPI_FUNC(void) _PyStackless_Init(void);

/* clean-up up at the end */

PyAPI_FUNC(void) PyStacklessEval_Fini(void);

PyAPI_FUNC(void) PyStackless_kill_tasks_with_stacks(void);

/* the special version of eval_frame */
PyAPI_FUNC(PyObject *) slp_eval_frame(struct _frame *f);

/* the frame dispatcher */
PyAPI_FUNC(PyObject *) slp_frame_dispatch(PyFrameObject *f);

/* the frame dispatcher for toplevel tasklets */
PyAPI_FUNC(PyObject *) slp_frame_dispatch_top(PyFrameObject *f);

/* the now exported eval_frame */
PyAPI_FUNC(PyObject *) PyEval_EvalFrame(struct _frame *f);

/* rebirth of software stack avoidance */

PyAPI_DATA(PyObject *) Py_UnwindToken;

/* frame cloning both needed in tasklets and generators */

PyAPI_FUNC(struct _frame *) slp_clone_frame(struct _frame *f);
PyAPI_FUNC(struct _frame *) slp_ensure_new_frame(struct _frame *f);

/* exposing the generator type */

PyAPI_DATA(PyTypeObject) PyGenerator_Type;

/********************************************************************
 *
 * This section defines/references stuff from stacklessmodule.c
 *
 ********************************************************************/

PyAPI_DATA(PyTypeObject) PyBaseFrame_Type;
PyAPI_DATA(PyTypeObject) PyCFrame_Type;

/* generic ops for chained objects */

/*  Tasklets are in doubly linked lists. We support
    deletion and insertion */

#define SLP_CHAIN_INSERT(__objtype, __chain, __task, __next, __prev) \
{ \
    __objtype *l, *r; \
	assert((__task)->__next == NULL); \
	assert((__task)->__prev == NULL); \
    if (*(__chain) == NULL) { \
        (__task)->__next = (__task)->__prev = (__task); \
        *(__chain) = (__task); \
    } \
    else { \
        /* insert at end */ \
        r = *(__chain); \
        l = r->__prev; \
        l->__next = r->__prev = (__task); \
        (__task)->__prev = l; \
        (__task)->__next = r; \
    } \
}

#define SLP_CHAIN_REMOVE(__objtype, __chain, __task, __next, __prev) \
{ \
    __objtype *l, *r; \
    if (*(__chain) == NULL) { \
        (__task) = NULL; \
    } \
    else { \
        /* remove current */ \
        (__task) = *(__chain); \
        l = (__task)->__prev; \
        r = (__task)->__next; \
        l->__next = r; \
        r->__prev = l; \
        *(__chain) = r; \
        if (*(__chain)==(__task)) \
            *(__chain) = NULL;  /* short circuit */ \
        (__task)->__prev = NULL; \
        (__task)->__next = NULL; \
    } \
}

/* operations on chains */

PyAPI_FUNC(void) slp_current_insert(PyTaskletObject *task);
PyAPI_FUNC(void) slp_current_insert_after(PyTaskletObject *task);
PyAPI_FUNC(PyTaskletObject *) slp_current_remove(void);
PyAPI_FUNC(void) slp_channel_insert(PyChannelObject *channel, PyTaskletObject *task, int dir);
PyAPI_FUNC(PyTaskletObject) * slp_channel_remove(PyChannelObject *channel, int dir);
PyAPI_FUNC(PyTaskletObject) * slp_channel_remove_specific(PyChannelObject *channel, int dir, PyTaskletObject *task);

/* tasklet operations */
PyAPI_FUNC(PyObject *) slp_tasklet_new(PyTypeObject *type, PyObject *args, PyObject *kwds);

PyAPI_FUNC(int) slp_schedule_task(PyTaskletObject *prev, PyTaskletObject *next);
PyAPI_FUNC(int) slp_schedule_nr_maybe(PyTaskletObject *prev, PyTaskletObject *next);

PyObject * slp_run_tasklet(PyFrameObject *f);
int initialize_main_and_current(PyFrameObject *f);

/* handy abbrevations */

PyObject * slp_type_error(const char *msg);
PyObject * slp_runtime_error(const char *msg);
PyObject * slp_value_error(const char *msg);
PyObject * slp_null_error(void);

/* this seems to be needed for gcc */

/* Define NULL pointer value */

#undef NULL
#ifdef  __cplusplus
#define NULL    0
#else
#define NULL    ((void *)0)
#endif

#define TYPE_ERROR(str, ret) (slp_type_error(str), ret)
#define RUNTIME_ERROR(str, ret) (slp_runtime_error(str), ret)
#define VALUE_ERROR(str, ret) (slp_value_error(str), ret)

PyBaseFrameObject * slp_baseframe_new(PyFrame_ExecFunc *exec, unsigned int linked, 
                                      unsigned int extra);
PyCFrameObject * slp_cframe_new(PyObject *func, PyObject *args, PyObject *kwds, unsigned int linked);

PyFrameObject * slp_get_frame(PyTaskletObject *task);
PyTaskletFlagStruc * slp_get_flags(PyTaskletObject *task);
int slp_return_wrapper(PyObject *retval);
int slp_int_wrapper(PyObject *retval);
int slp_current_wrapper( int(*func)(PyTaskletObject*), PyTaskletObject *task);
int slp_revive_main(void);

/* debugging/monitoring */

typedef void (slp_schedule_hook_func) (PyTaskletObject *from, PyTaskletObject *to);
PyAPI_DATA(slp_schedule_hook_func) *_slp_schedule_fasthook;
PyAPI_DATA(PyObject* ) _slp_schedule_hook;
void slp_schedule_callback(PyTaskletObject *prev, PyTaskletObject *next);

#ifdef DEEP_DEBUG
void _deep_debug(PyTaskletObject *prev, PyTaskletObject *target);
#endif

#include "stackless_api.h"

#endif

#ifdef __cplusplus
}
#endif

#endif
